#==============================================================================

# Copyright (c) 2016-11-21, Laurenz Wiskott, http://www.ini.rub.de/PEOPLE/wiskott/
# License: BSD 2-Clause, https://opensource.org/licenses/BSD-2-Clause, see the end

#==============================================================================

import numpy as np
import math as m
from scipy.stats import multivariate_normal

class Gaussian:

    DIM_DEFAULT = 2
    MEAN_STD_DEFAULT = 2
    AXES_DEFAULT = (0,1)
    NUM_DEFAULT = 30
    POINT_SIZE = 3
    COLOR_DEFAULT = 'k'
    LINESTYLE_DEFAULT = '-'
    LINEWIDTH_DEFAULT = 10
    
    def __init__(self, dim=DIM_DEFAULT, mean_std=MEAN_STD_DEFAULT, mean='random',
                 trafo='random', num=NUM_DEFAULT, color=COLOR_DEFAULT, linestyle=LINESTYLE_DEFAULT):
        """Initialize Gaussian with either given or random mean and trafo.

        dim (int): dimension of the Gaussian
        mean_std (float): standard deviation of the mean distribution if drawn randomly from a Gaussian
        mean (array of floats with shape dim): mean, either vector or 'random' if components shall be drawn from a Gaussian
        trafo (array of floats with shape dim x dim): transformation that would turn a normalized Gaussian to the given Gaussian, 
            either matrix or 'random' if components shall be drawn from a Gaussian
        num (int): weight of the Gaussian, as if num data points are typically drawn from it
        """

        self.dim = dim
        self.num = num
        if mean == 'random':
            self.mean = mean_std*np.random.normal(0,1,(dim))
        else:
            self.mean = np.array(mean)
        if trafo == 'random':
            pc_var_min = 0
            while pc_var_min < 0.5:
                self.trafo(np.random.normal(0,1,(dim,dim)))
                pc_var, pcVec = np.linalg.eig(self.cov())
                pc_var_min = np.nanmin(pc_var)
        else:
            self.trafo(np.array(trafo))
        self.color = color
        self.linestyle = linestyle


    def trafo(self, trafo=None):

        if trafo is not None:
            self._trafo = trafo
            self._cov = np.dot(self._trafo, self._trafo.transpose())

        return self._trafo


    def cov(self, cov=None):

        if cov is not None:
            self._cov = cov
            try:
                pc_var, pcVec = np.linalg.eig(self.cov())
            except np.linalg.LinAlgError:
                print("Exception: Matrix is probably singular in 'Gaussian.cov'!")
                raise
            self._trafo = np.dot(np.dot(pcVec,np.diag(np.sqrt(pc_var))),pcVec.transpose())

        return self._cov


    def level_line_path(self, std=1, points=36, axes=(0,1)):
        """Compute a level line path (ellipse) for a Gaussian for plotting in 2D.
        
        std (float): standard deviation level at which the ellipse should be drawn
        points (int): number of points of the path
        axes(tuple of 2 ints): two axes that define the plain for which the ellipse is drawn
        """

        mean_2d = self.mean[np.ix_(axes)]
        trafo_2d = self._trafo[np.ix_(axes,axes)]
        return mean_2d.reshape((2,1)) \
            + np.dot(trafo_2d,np.array([[m.cos(phi*m.pi/180),m.sin(phi*m.pi/180)] for phi in range(0,361,360//points)]).transpose())

    def ax_plot(self, ax, axes=AXES_DEFAULT, linewidth=LINEWIDTH_DEFAULT):

        if self.mean is not None and self._trafo is not None:
            path = self.level_line_path(self)
            ax.plot(path[0], path[1], color=self.color, linestyle=self.linestyle, linewidth=linewidth)


    def pd_point(self, point):

        point = np.array(point)

        try:
            return multivariate_normal.pdf(point, mean=self.mean, cov=self.cov())
        except np.linalg.LinAlgError:
            print("Exception: Matrix is probably singular in 'Gaussian.pd_point'!")
            raise


    def pd_points(self, points):

        points = np.array(points)

        try:
            return multivariate_normal.pdf(points, mean=self.mean, cov=self.cov())
        except np.linalg.LinAlgError:
            print("Exception: Matrix is probably singular in 'Gaussian.pd_points'!")
            raise


    def sample_points(self, num=NUM_DEFAULT):

        return multivariate_normal.rvs(mean=self.mean, cov=self.cov(), size=num)


#==============================================================================
#==============================================================================

# BSD 2-Clause License

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#==============================================================================

