# ==============================================================================

# Copyright (c) 2016-11-21, Laurenz Wiskott, http://www.ini.rub.de/PEOPLE/wiskott/
# License: BSD 2-Clause, https://opensource.org/licenses/BSD-2-Clause, see the end

# ==============================================================================

import numpy as np
import matplotlib.pyplot as plt
import time

from mixture_of_gaussian_clusters import MixtureOfGaussianClusters
from mixture_of_gaussians import MixtureOfGaussians

# Process parameters.
VISUALIZE_PROCESS = True  # Visualize the process of clustering (or not).
SLEEP_TIME = 2.  # Time to sleep after each visualized process.
EVALUATE_PERFORMANCE = True  # Visualize the results of clustering (or not).

# Data parameters.
DIM = 2  # It does not make much sense to use other values than 2 here.
NUM_CLUSTERS_MIN = 1  # Minimum number of clusters in the data.
NUM_CLUSTERS_MAX = 6  # Maximum number of clusters in the data.
NUM_POINTS_MIN = 20  # Minimum number of points in a cluster.
NUM_POINTS_MAX = 100  # Maximum number of points in a cluster.
MEAN_STD = 10  # Mean standard deviation of the cluster centers.

# Optimization parameters.
NUM_TEST = 100  # Assumed number of data points in the Gaussian components.
NUM_REPETITIONS_TEST_PROCESS = [1, 2, 3, 4, 5, 6]  # Number of trials for different numbers of Gaussian components.
NUM_REPETITIONS_TEST_RESULTS = [1, 3, 6, 10, 15, 21]  # Number of trials for different numbers of Gaussian components.
SESSIONS = 100  # Number of times the whole optimization process is run for different data.


def repetitions_list_template(num_repetitions_test):
    """Provides a suitably dimensioned list of None values as a template to be filled later with some data."""

    return [[None
             for trial_idx in range(num_repetitions_test[nc_idx])]
            for nc_idx in range(len(num_repetitions_test))]


def new_mixture_of_clusters(init_num_clusters=None):
    """Initialize mixture of Gaussian clusters."""

    # Set number of clusters.
    if init_num_clusters is None:
        num_clusters = np.random.random_integers(NUM_CLUSTERS_MIN, NUM_CLUSTERS_MAX)
    else:
        num_clusters = NUM_CLUSTERS_MIN + init_num_clusters%(NUM_CLUSTERS_MAX-NUM_CLUSTERS_MIN+1)
    # Set number of data points in the clusters.
    num = np.random.random_integers(NUM_POINTS_MIN, NUM_POINTS_MAX, num_clusters) 

    # Return mixture of clusters.
    return MixtureOfGaussianClusters(dim=DIM, mean_std=MEAN_STD, num_components=num_clusters, num=num)


def plot_mixture_model(fig, ax, mg, mc, color_clusters=True):
    """Plot a mixture model along with the data fitted."""

    # Determine plot ranges.
    points = mc.points()
    pnt_x_min, pnt_x_max = np.min(points[:, 0]), np.max(points[:, 0])
    pnt_y_min, pnt_y_max = np.min(points[:, 1]), np.max(points[:, 1])
    pnt_x_center = (pnt_x_min + pnt_x_max) // 2
    pnt_y_center = (pnt_y_min + pnt_y_max) // 2
    pnt_range_max = np.max([pnt_x_max-pnt_x_min, pnt_y_max-pnt_y_min])
    x_lim_min, x_lim_max = pnt_x_center - pnt_range_max/1.8, pnt_x_center + pnt_range_max/1.8, 
    y_lim_min, y_lim_max = pnt_y_center - pnt_range_max/1.8, pnt_y_center + pnt_range_max/1.8, 

    # Plot figure.
    ax.clear()
    ax.set(xlim=[x_lim_min, x_lim_max], ylim=[y_lim_min, y_lim_max], title='Data and Gaussians', 
           xlabel='x', ylabel='y')
    if color_clusters:
        mc.ax_plot(ax)
    mg.ax_plot(ax, show_target=not color_clusters)
    fig.canvas.draw()


def fit_mixture_models(points, num_repetitions_test=[1, 2, 3, 4, 5, 6], visualize=True, color_clusters=False):
    """Fit a mixture model to some data points."""

    # Prepare figure to show the process of optimizing individual Gaussian mixture models.
    if visualize:
        fig_process = plt.figure(figsize=(12, 12))
        ax_process = plt.subplot2grid((1, 1), (0, 0))
        plt.show(block=False)

    # Iterate a number of Gaussian mixture models for different numbers of Gaussian components.
    iterated_mgs = repetitions_list_template(num_repetitions_test)
    for nc_idx in range(len(num_repetitions_test)):
        num_components = nc_idx + 1
        print()
        for trial_idx in range(num_repetitions_test[nc_idx]):
            successful_fit = False
            while not successful_fit:
                print("Trial", trial_idx+1, "/", num_repetitions_test[nc_idx], "with",
                      num_components, "components.")
                mg = MixtureOfGaussians(dim=DIM, mean_std=MEAN_STD,
                                        num_components=num_components, num=NUM_TEST, target_points=points)
                for counter2 in range(1000):
                    try:
                        delta_mean, delta_cov = mg.iterate_fit(points)
                        # This is a dummy call to check whether it works (the matrix might be singular).
                        pds = mg.pd_points(points)
                    except np.linalg.LinAlgError:
                        print("Exception: Matrix is probably singular in 'main', start new optimization!")
                        mg = None
                        successful_fit = True
                        break

                    # SOLUTION:
                    # Discard optimizations where one Gaussian component zooms in on too few data points.
                    if np.min(mg._prob_component)*np.sum(mc.num) < 15:
                        print("Warning: One component focuses on fewer than 15 data points, start new optimization!")
                        mg = None
                        successful_fit = True
                        break

                    if visualize:
                        plot_mixture_model(fig_process, ax_process, mg, mc, color_clusters=color_clusters)
                    if delta_mean < 0.0001 and delta_cov < 0.0001:
                        successful_fit = True
                        break
            iterated_mgs[nc_idx][trial_idx] = mg
            if visualize:
                time.sleep(SLEEP_TIME)

    return iterated_mgs


if VISUALIZE_PROCESS:
    # For one mixture of clusters show a set of fitting processes in detail.

    # Initialize mixture of clusters.
    mc = new_mixture_of_clusters()

    # Fit mixture of Gaussians.
    iterated_mgs = fit_mixture_models(mc.points(), NUM_REPETITIONS_TEST_PROCESS, visualize=True, color_clusters=True)

    print()
    print("====================================================================================")
    print("====================================================================================")
    print()

# ==============================================================================

# EXERCISE 1: Add the missing code in file 'mixture.py' to make the fitting
# process work.  Look for the EXERCISE keyword in file 'mixture.py'.

# SOLUTION:
# Look for the SOLUTION keyword in file 'mixture.py'.

# ------------------------------------------------------------------------------

# EXERCISE 2: The function 'fit_mixture_models' generates a whole list of fits
# with different numbers of Gaussian components and several runs for each
# number as determined by the second argument (num_repetitions_test).  Try to
# figure out an automatic way of determining which of these many fits
# estimates the number of clusters in the data best.  Do some statistics to
# determine the percentage of correct estimates.  Ignore the complication that
# sometimes two clusters sit right on top of each other and cannot be
# separated.  Always use the nominal number of clusters as a reference.

# SOLUTION:


def select_best_mixture_model(log_likelihood_weight, log_likelihoods,
                              partition_coefficient_index_weight, partition_coefficient_indexs):

    best_fitness_value = None
    for nc_idx in range(len(NUM_REPETITIONS_TEST_RESULTS)):
        num_components = nc_idx + 1
        for trial_idx in range(NUM_REPETITIONS_TEST_RESULTS[nc_idx]):
            trial = trial_idx + 1
            if log_likelihoods[nc_idx][trial_idx] is not None \
               and partition_coefficient_indexs[nc_idx][trial_idx] is not None:
                this_fitness_value = log_likelihood_weight*log_likelihoods[nc_idx][trial_idx] \
                                     + partition_coefficient_index_weight*partition_coefficient_indexs[nc_idx][trial_idx]
                if best_fitness_value == None or this_fitness_value > best_fitness_value:
                    best_num_components = num_components
                    best_trial = trial
                    best_fitness_value = this_fitness_value

    return best_num_components, best_trial, best_fitness_value


if EVALUATE_PERFORMANCE:
    # Run the whole session repeatedly to get some statistics.
    # Prepare figure to visualize best fitting model.
    fig_optimum = plt.figure(figsize=(12, 12))
    ax_optimum = plt.subplot2grid((1, 1), (0, 0))
    # Prepare figure to visualize statistical features.
    fig_statistics = plt.figure(figsize=(12, 12))
    ax_statistics = plt.subplot2grid((1, 1), (0, 0))
    # Prepare history lists.
    log_likelihoods_history = []
    partition_coefficient_indexs_history = []
    mc_num_components_history = []
    successes = 0
    best_num_components_estimates = []
    for session in range(SESSIONS):
        print("Session", session)
        ax_statistics.clear()
        ax_statistics.set(title='Statistics over several runs (black=wrong, pink=correct)', xlabel='log likelihood', ylabel='partition coefficient index')

        # Set seed to get reproducable results in detail.
        np.random.seed(session)

        # Initialize mixture of clusters.
        mc = new_mixture_of_clusters(init_num_clusters=session)
        mc_num_components_history.append(mc.num_components)

        # Fit mixture of Gaussians.
        iterated_mgs = fit_mixture_models(mc.points(), NUM_REPETITIONS_TEST_RESULTS, visualize=False)

        # Calculate statistical features of the fits.
        log_likelihoods = repetitions_list_template(NUM_REPETITIONS_TEST_RESULTS)
        partition_coefficient_indexs = repetitions_list_template(NUM_REPETITIONS_TEST_RESULTS)
        correct_nums = repetitions_list_template(NUM_REPETITIONS_TEST_RESULTS)
        print()
        print("Num clusters, num components, trial, log likelihood, partition coefficient index:")
        for nc_idx in range(len(NUM_REPETITIONS_TEST_RESULTS)):
            num_components = nc_idx + 1
            for trial_idx in range(NUM_REPETITIONS_TEST_RESULTS[nc_idx]):
                mg = iterated_mgs[nc_idx][trial_idx]
                if mg is None:
                    continue
                pds = mg.pd_points(mc.points())
                log_likelihoods[nc_idx][trial_idx] = np.average(np.log(pds))
                partition_coefficient_indexs[nc_idx][trial_idx] = \
                    np.sum(mg.prob_component_given_point**2) \
                    / len(mg.target_points)
                if num_components == mc.num_components:
                    correct_nums[nc_idx][trial_idx] = 1
                else:
                    correct_nums[nc_idx][trial_idx] = 0
                print(mc.num_components, num_components, trial_idx+1, \
                    log_likelihoods[nc_idx][trial_idx], \
                    partition_coefficient_indexs[nc_idx][trial_idx])
                ax_statistics.scatter([log_likelihoods[nc_idx][trial_idx]],
                    [partition_coefficient_indexs[nc_idx][trial_idx]],
                    color=['k', 'm'][correct_nums[nc_idx][trial_idx]])
                fig_statistics.canvas.draw()
        log_likelihoods_history.append(log_likelihoods)
        partition_coefficient_indexs_history.append(partition_coefficient_indexs)

        # Select best fitting model.
        best_num_components, best_trial, best_fitness_value = \
            select_best_mixture_model(0.5, log_likelihoods, 0.5, partition_coefficient_indexs)

        # Visualize best fitting model.
        plot_mixture_model(fig_optimum, ax_optimum, iterated_mgs[best_num_components-1][best_trial-1],
                           mc, color_clusters=True)
        plt.show(block=False)

        best_num_components_estimates.append([mc.num_components, best_num_components])
        print("\n", mc.num_components, "components estimated as", best_num_components,
              "with fitness", best_fitness_value, end="")
        if mc.num_components == best_num_components:
            print("-> CORRECT!")
            successes += 1
        else:
            print("-> WRONG!")
        print()
        print("====================================================================================")
        print()

    best_success_rate = float(successes)/SESSIONS


    def print_results():

        print()
        print("NUM_CLUSTERS_MIN:", NUM_CLUSTERS_MIN)
        print("NUM_CLUSTERS_MAX:", NUM_CLUSTERS_MAX)
        print("NUM_POINTS_MIN:", NUM_POINTS_MIN)
        print("NUM_POINTS_MAX:", NUM_POINTS_MAX)
        print("MEAN_STD:", MEAN_STD)
        print("NUM_TEST:", NUM_TEST)
        print("NUM_REPETITIONS_TEST_RESULTS:", NUM_REPETITIONS_TEST_RESULTS)
        print("SESSIONS:", SESSIONS)
        print()
        print("Success rat:", best_success_rate)
        print()
        best_num_components_estimate_matrix = np.zeros([NUM_CLUSTERS_MAX, len(NUM_REPETITIONS_TEST_RESULTS)])
        for best_num_components_estimate in np.array(best_num_components_estimates):
            best_num_components_estimate_matrix[best_num_components_estimate[0]-1, best_num_components_estimate[1]-1] += 1
        print("Best fit matrix (vertical = true number of clusters, horizontal = estimated number of cluster)\n", best_num_components_estimate_matrix)
        print()

    print_results()


    # Search for best value of 'log_likelihood_weight'.
    best_log_likelihood_weight = None
    best_success_rate = None
    best_num_components_estimates = None
    print()
    print("Log likelihood weight, success rate")
    # Go through a grid of values and test success rate.
    for log_likelihood_weight in np.arange(0, 1.1, 0.1):
        successes = 0
        num_components_estimates = []
        for session in range(SESSIONS):

            # Select best fitting model.
            best_num_components, best_trial, best_fitness_value = \
                select_best_mixture_model(log_likelihood_weight, log_likelihoods_history[session], 
                    1.-log_likelihood_weight, partition_coefficient_indexs_history[session])

            num_components_estimates.append([mc_num_components_history[session], best_num_components])
            if mc_num_components_history[session] == best_num_components:
                successes += 1

        success_rate = float(successes)/SESSIONS
        print(log_likelihood_weight, success_rate)
        if best_success_rate is None or success_rate > best_success_rate:
            best_success_rate = success_rate
            best_log_likelihood_weight = log_likelihood_weight
            best_num_components_estimates = num_components_estimates

    print("Best log likelihood weight and success rate:", best_log_likelihood_weight, best_success_rate)

    print_results()

#==============================================================================
#==============================================================================

# BSD 2-Clause License

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#==============================================================================

